@@ -41,8 +41,6 @@
 //!
 //! These defined server will be used with a load balancing algorithm.
 
-#[cfg(any(unix, target_os = "android", feature = "local-flow-stat"))]
-use std::path::PathBuf;
 use std::{
     convert::{From, Infallible},
     default::Default,
@@ -51,18 +49,20 @@ use std::{
     io::Read,
     net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr, SocketAddrV4, SocketAddrV6},
     option::Option,
-    path::Path,
+    path::{Path, PathBuf},
     str::FromStr,
     string::ToString,
     time::Duration,
 };
 
 use cfg_if::cfg_if;
+#[cfg(feature = "local-tun")]
+use ipnet::IpNet;
 use serde::{Deserialize, Serialize};
 #[cfg(any(feature = "local-tunnel", feature = "local-dns"))]
 use shadowsocks::relay::socks5::Address;
 use shadowsocks::{
-    config::{ManagerAddr, Mode, ServerAddr, ServerConfig, ServerWeight},
+    config::{ManagerAddr, Mode, ReplayAttackPolicy, ServerAddr, ServerConfig, ServerWeight},
     crypto::v1::CipherKind,
     plugin::PluginConfig,
 };
@@ -81,6 +81,26 @@ enum SSDnsConfig {
     TrustDns(ResolverConfig),
 }
 
+#[derive(Serialize, Deserialize, Debug, Default)]
+struct SSSecurityConfig {
+    #[serde(skip_serializing_if = "Option::is_none")]
+    replay_attack: Option<SSSecurityReplayAttackConfig>,
+}
+
+#[derive(Serialize, Deserialize, Debug, Default)]
+struct SSSecurityReplayAttackConfig {
+    #[serde(skip_serializing_if = "Option::is_none")]
+    policy: Option<String>,
+}
+
+#[derive(Serialize, Deserialize, Debug, Default)]
+struct SSBalancerConfig {
+    #[serde(skip_serializing_if = "Option::is_none")]
+    max_server_rtt: Option<u64>,
+    #[serde(skip_serializing_if = "Option::is_none")]
+    check_interval: Option<u64>,
+}
+
 #[derive(Serialize, Deserialize, Debug, Default)]
 struct SSConfig {
     #[serde(skip_serializing_if = "Option::is_none")]
@@ -113,7 +133,7 @@ struct SSConfig {
     udp_timeout: Option<u64>,
     #[serde(skip_serializing_if = "Option::is_none")]
     udp_max_associations: Option<usize>,
-    #[serde(skip_serializing_if = "Option::is_none")]
+    #[serde(skip_serializing_if = "Option::is_none", alias = "shadowsocks")]
     servers: Option<Vec<SSServerExtConfig>>,
     #[serde(skip_serializing_if = "Option::is_none")]
     locals: Option<Vec<SSLocalExtConfig>>,
@@ -132,12 +152,18 @@ struct SSConfig {
     ipv6_first: Option<bool>,
     #[serde(skip_serializing_if = "Option::is_none")]
     fast_open: Option<bool>,
+    #[serde(skip_serializing_if = "Option::is_none")]
+    security: Option<SSSecurityConfig>,
+    #[serde(skip_serializing_if = "Option::is_none")]
+    balancer: Option<SSBalancerConfig>,
 }
 
 #[derive(Serialize, Deserialize, Debug, Default)]
 struct SSLocalExtConfig {
+    #[serde(skip_serializing_if = "Option::is_none")]
     local_address: Option<String>,
-    local_port: u16,
+    #[serde(skip_serializing_if = "Option::is_none")]
+    local_port: Option<u16>,
 
     #[serde(skip_serializing_if = "Option::is_none")]
     disabled: Option<bool>,
@@ -188,6 +214,14 @@ struct SSLocalExtConfig {
     #[cfg(feature = "local-tunnel")]
     #[serde(skip_serializing_if = "Option::is_none")]
     forward_port: Option<u16>,
+
+    /// Tun
+    #[cfg(feature = "local-tun")]
+    #[serde(skip_serializing_if = "Option::is_none")]
+    tun_interface_name: Option<String>,
+    #[cfg(feature = "local-tun")]
+    #[serde(skip_serializing_if = "Option::is_none")]
+    tun_interface_address: Option<String>,
 }
 
 #[derive(Serialize, Deserialize, Debug)]
@@ -211,7 +245,7 @@ struct SSServerExtConfig {
     plugin_args: Option<Vec<String>>,
     #[serde(skip_serializing_if = "Option::is_none")]
     timeout: Option<u64>,
-    #[serde(skip_serializing_if = "Option::is_none")]
+    #[serde(skip_serializing_if = "Option::is_none", alias = "name")]
     remarks: Option<String>,
     #[serde(skip_serializing_if = "Option::is_none")]
     id: Option<String>,
@@ -507,6 +541,50 @@ impl FromStr for ManagerServerHost {
     }
 }
 
+/// Mode of Manager's server
+#[derive(Debug, Clone, Copy, Eq, PartialEq)]
+pub enum ManagerServerMode {
+    /// Run shadowsocks server in the same process of manager
+    Builtin,
+
+    /// Run shadowsocks server in standalone (process) mode
+    #[cfg(unix)]
+    Standalone,
+}
+
+impl Default for ManagerServerMode {
+    fn default() -> ManagerServerMode {
+        ManagerServerMode::Builtin
+    }
+}
+
+/// Parsing ManagerServerMode error
+#[derive(Debug, Clone, Copy)]
+pub struct ManagerServerModeError;
+
+impl FromStr for ManagerServerMode {
+    type Err = ManagerServerModeError;
+
+    fn from_str(s: &str) -> Result<ManagerServerMode, Self::Err> {
+        match s {
+            "builtin" => Ok(ManagerServerMode::Builtin),
+            #[cfg(unix)]
+            "standalone" => Ok(ManagerServerMode::Standalone),
+            _ => Err(ManagerServerModeError),
+        }
+    }
+}
+
+impl Display for ManagerServerMode {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match *self {
+            ManagerServerMode::Builtin => f.write_str("builtin"),
+            #[cfg(unix)]
+            ManagerServerMode::Standalone => f.write_str("standalone"),
+        }
+    }
+}
+
 /// Configuration for Manager
 #[derive(Clone, Debug)]
 pub struct ManagerConfig {
@@ -514,6 +592,8 @@ pub struct ManagerConfig {
     pub addr: ManagerAddr,
     /// Manager's default method
     pub method: Option<CipherKind>,
+    /// Manager's default plugin
+    pub plugin: Option<PluginConfig>,
     /// Timeout for TCP connections, setting to manager's created servers
     pub timeout: Option<Duration>,
     /// IP/Host for servers to bind (inbound)
@@ -522,6 +602,14 @@ pub struct ManagerConfig {
     pub server_host: ManagerServerHost,
     /// Server's mode
     pub mode: Mode,
+    /// Server's running mode
+    pub server_mode: ManagerServerMode,
+    /// Server's command if running in Standalone mode
+    #[cfg(unix)]
+    pub server_program: String,
+    /// Server's working directory if running in Standalone mode
+    #[cfg(unix)]
+    pub server_working_directory: PathBuf,
 }
 
 impl ManagerConfig {
@@ -530,9 +618,18 @@ impl ManagerConfig {
         ManagerConfig {
             addr,
             method: None,
+            plugin: None,
             timeout: None,
             server_host: ManagerServerHost::default(),
             mode: Mode::TcpOnly,
+            server_mode: ManagerServerMode::Builtin,
+            #[cfg(unix)]
+            server_program: "ssserver".to_owned(),
+            #[cfg(unix)]
+            server_working_directory: match std::env::current_dir() {
+                Ok(d) => d,
+                Err(..) => "/tmp/shadowsocks-manager".into(),
+            },
         }
     }
 }
@@ -549,6 +646,8 @@ pub enum ProtocolType {
     Redir,
     #[cfg(feature = "local-dns")]
     Dns,
+    #[cfg(feature = "local-tun")]
+    Tun,
 }
 
 impl Default for ProtocolType {
@@ -570,6 +669,8 @@ impl ProtocolType {
             ProtocolType::Redir => "redir",
             #[cfg(feature = "local-dns")]
             ProtocolType::Dns => "dns",
+            #[cfg(feature = "local-tun")]
+            ProtocolType::Tun => "tun",
         }
     }
 
@@ -585,6 +686,8 @@ impl ProtocolType {
             "redir",
             #[cfg(feature = "local-dns")]
             "dns",
+            #[cfg(feature = "local-tun")]
+            "tun",
         ]
     }
 }
@@ -607,6 +710,8 @@ impl FromStr for ProtocolType {
             "redir" => Ok(ProtocolType::Redir),
             #[cfg(feature = "local-dns")]
             "dns" => Ok(ProtocolType::Dns),
+            #[cfg(feature = "local-tun")]
+            "tun" => Ok(ProtocolType::Tun),
             _ => Err(ProtocolTypeError),
         }
     }
@@ -615,7 +720,9 @@ impl FromStr for ProtocolType {
 /// Local server configuration
 #[derive(Clone, Debug)]
 pub struct LocalConfig {
-    pub addr: ServerAddr,
+    /// Listen address for local servers
+    pub addr: Option<ServerAddr>,
+
     pub protocol: ProtocolType,
 
     /// Mode
@@ -648,13 +755,30 @@ pub struct LocalConfig {
     /// Sending DNS query through proxy to this address
     #[cfg(feature = "local-dns")]
     pub remote_dns_addr: Option<Address>,
+
+    /// Tun interface's name
+    ///
+    /// Linux: eth0, eth1, ...
+    /// macOS: utun0, utun1, ...
+    #[cfg(feature = "local-tun")]
+    pub tun_interface_name: Option<String>,
+    /// Tun interface's address and netmask
+    #[cfg(feature = "local-tun")]
+    pub tun_interface_address: Option<IpNet>,
+    /// Tun interface's file descriptor
+    #[cfg(all(feature = "local-tun", unix))]
+    pub tun_device_fd: Option<std::os::unix::io::RawFd>,
+    /// Tun interface's file descriptor read from this Unix Domain Socket
+    #[cfg(all(feature = "local-tun", unix))]
+    pub tun_device_fd_from_path: Option<PathBuf>,
 }
 
 impl LocalConfig {
     /// Create a new `LocalConfig`
-    pub fn new(addr: ServerAddr, protocol: ProtocolType) -> LocalConfig {
+    pub fn new(protocol: ProtocolType) -> LocalConfig {
         LocalConfig {
-            addr,
+            addr: None,
+
             protocol,
 
             mode: Mode::TcpOnly,
@@ -672,10 +796,38 @@ impl LocalConfig {
             local_dns_addr: None,
             #[cfg(feature = "local-dns")]
             remote_dns_addr: None,
+
+            #[cfg(feature = "local-tun")]
+            tun_interface_name: None,
+            #[cfg(feature = "local-tun")]
+            tun_interface_address: None,
+            #[cfg(all(feature = "local-tun", unix))]
+            tun_device_fd: None,
+            #[cfg(all(feature = "local-tun", unix))]
+            tun_device_fd_from_path: None,
         }
     }
 
+    /// Create a new `LocalConfig` with listen address
+    pub fn new_with_addr(addr: ServerAddr, protocol: ProtocolType) -> LocalConfig {
+        let mut config = LocalConfig::new(protocol);
+        config.addr = Some(addr);
+        config
+    }
+
     fn check_integrity(&self) -> Result<(), Error> {
+        match self.protocol {
+            #[cfg(feature = "local-tun")]
+            ProtocolType::Tun => {}
+
+            _ => {
+                if self.addr.is_none() {
+                    let err = Error::new(ErrorKind::MissingField, "missing `addr` in configuration", None);
+                    return Err(err);
+                }
+            }
+        }
+
         match self.protocol {
             #[cfg(feature = "local-dns")]
             ProtocolType::Dns => {
@@ -696,6 +848,14 @@ impl LocalConfig {
                 }
             }
 
+            #[cfg(feature = "local-http")]
+            ProtocolType::Http => {
+                if !self.mode.enable_tcp() {
+                    let err = Error::new(ErrorKind::Invalid, "TCP mode have to be enabled for http", None);
+                    return Err(err);
+                }
+            }
+
             _ => {}
         }
 
@@ -742,13 +902,31 @@ impl Default for DnsConfig {
     }
 }
 
+/// Security Config
+#[derive(Clone, Debug, Default)]
+pub struct SecurityConfig {
+    pub replay_attack: SecurityReplayAttackConfig,
+}
+
+#[derive(Clone, Debug, Default)]
+pub struct SecurityReplayAttackConfig {
+    pub policy: ReplayAttackPolicy,
+}
+
+/// Balancer Config
+#[derive(Clone, Debug, Default)]
+pub struct BalancerConfig {
+    /// MAX rtt of servers, which is the timeout duration of each check requests
+    pub max_server_rtt: Option<Duration>,
+    /// Interval between each checking
+    pub check_interval: Option<Duration>,
+}
+
 /// Configuration
 #[derive(Clone, Debug)]
 pub struct Config {
     /// Remote ShadowSocks server configurations
     pub server: Vec<ServerConfig>,
-    /// Local server's bind address, or ShadowSocks server's outbound address
-    pub local_addr: Option<IpAddr>,
     /// Local server configuration
     pub local: Vec<LocalConfig>,
 
@@ -785,9 +963,10 @@ pub struct Config {
     /// Set `SO_MARK` socket option for outbound sockets
     #[cfg(any(target_os = "linux", target_os = "android"))]
     pub outbound_fwmark: Option<u32>,
-    /// Set `SO_BINDTODEVICE` socket option for outbound sockets
-    #[cfg(any(target_os = "linux", target_os = "android", target_os = "macos", target_os = "ios"))]
+    /// Set `SO_BINDTODEVICE` (Linux), `IP_BOUND_IF` (BSD), `IP_UNICAST_IF` (Windows) socket option for outbound sockets
     pub outbound_bind_interface: Option<String>,
+    /// Outbound sockets will `bind` to this address
+    pub outbound_bind_addr: Option<IpAddr>,
     /// Path to protect callback unix address, only for Android
     #[cfg(target_os = "android")]
     pub outbound_vpn_protect_path: Option<PathBuf>,
@@ -818,6 +997,16 @@ pub struct Config {
     /// Flow statistic report Unix socket path (only for Android)
     #[cfg(feature = "local-flow-stat")]
     pub stat_path: Option<PathBuf>,
+
+    /// Replay attack policy
+    pub security: SecurityConfig,
+
+    /// Balancer config of local server
+    pub balancer: BalancerConfig,
+
+    /// Configuration file path, the actual path of the configuration.
+    /// This is normally for auto-reloading if implementation supports.
+    pub config_path: Option<PathBuf>,
 }
 
 /// Configuration parsing error kind
@@ -884,7 +1073,6 @@ impl Config {
     pub fn new(config_type: ConfigType) -> Config {
         Config {
             server: Vec::new(),
-            local_addr: None,
             local: Vec::new(),
 
             dns: DnsConfig::default(),
@@ -899,8 +1087,8 @@ impl Config {
 
             #[cfg(any(target_os = "linux", target_os = "android"))]
             outbound_fwmark: None,
-            #[cfg(any(target_os = "linux", target_os = "android", target_os = "macos", target_os = "ios"))]
             outbound_bind_interface: None,
+            outbound_bind_addr: None,
             #[cfg(target_os = "android")]
             outbound_vpn_protect_path: None,
 
@@ -920,6 +1108,12 @@ impl Config {
 
             #[cfg(feature = "local-flow-stat")]
             stat_path: None,
+
+            security: SecurityConfig::default(),
+
+            balancer: BalancerConfig::default(),
+
+            config_path: None,
         }
     }
 
@@ -989,7 +1183,8 @@ impl Config {
                         get_local_address(config.local_address, local_port, config.ipv6_first.unwrap_or(false));
 
                     // shadowsocks uses SOCKS5 by default
-                    let mut local_config = LocalConfig::new(local_addr, ProtocolType::Socks);
+                    let mut local_config = LocalConfig::new(ProtocolType::Socks);
+                    local_config.addr = Some(local_addr);
                     local_config.mode = global_mode;
                     local_config.protocol = match config.protocol {
                         None => ProtocolType::Socks,
@@ -1016,17 +1211,6 @@ impl Config {
                             continue;
                         }
 
-                        if local.local_port == 0 {
-                            let err = Error::new(ErrorKind::Malformed, "`local_port` cannot be 0", None);
-                            return Err(err);
-                        }
-
-                        let local_addr = get_local_address(
-                            local.local_address,
-                            local.local_port,
-                            config.ipv6_first.unwrap_or(false),
-                        );
-
                         let protocol = match local.protocol {
                             None => ProtocolType::Socks,
                             Some(p) => match p.parse::<ProtocolType>() {
@@ -1042,7 +1226,21 @@ impl Config {
                             },
                         };
 
-                        let mut local_config = LocalConfig::new(local_addr, protocol);
+                        let mut local_config = LocalConfig::new(protocol);
+
+                        if let Some(local_port) = local.local_port {
+                            if local_port == 0 {
+                                let err = Error::new(ErrorKind::Malformed, "`local_port` cannot be 0", None);
+                                return Err(err);
+                            }
+
+                            let local_addr =
+                                get_local_address(local.local_address, local_port, config.ipv6_first.unwrap_or(false));
+                            local_config.addr = Some(local_addr);
+                        } else if local.local_address.is_some() {
+                            let err = Error::new(ErrorKind::Malformed, "missing `local_port`", None);
+                            return Err(err);
+                        }
 
                         if let Some(local_udp_port) = local.local_udp_port {
                             if local_udp_port == 0 {
@@ -1142,24 +1340,31 @@ impl Config {
                             });
                         }
 
+                        #[cfg(feature = "local-tun")]
+                        if let Some(tun_interface_address) = local.tun_interface_address {
+                            match tun_interface_address.parse::<IpNet>() {
+                                Ok(addr) => local_config.tun_interface_address = Some(addr),
+                                Err(..) => {
+                                    let err = Error::new(ErrorKind::Malformed, "`tun_interface_address` invalid", None);
+                                    return Err(err);
+                                }
+                            }
+                        }
+
+                        #[cfg(feature = "local-tun")]
+                        if let Some(tun_interface_name) = local.tun_interface_name {
+                            local_config.tun_interface_name = Some(tun_interface_name);
+                        }
+
                         nconfig.local.push(local_config);
                     }
                 }
             }
             ConfigType::Server | ConfigType::Manager => {
+                // NOTE: IGNORED.
                 // servers only uses `local_address` for binding outbound interfaces
-
-                if let Some(local_address) = config.local_address {
-                    match local_address.parse::<IpAddr>() {
-                        Ok(ip) => {
-                            nconfig.local_addr = Some(ip);
-                        }
-                        Err(..) => {
-                            let err = Error::new(ErrorKind::Malformed, "`local_address` invalid", None);
-                            return Err(err);
-                        }
-                    }
-                }
+                //
+                // This behavior causes lots of confusion. use outbound_bind_addr instead
             }
         }
 
@@ -1190,14 +1395,14 @@ impl Config {
                 let mut nsvr = ServerConfig::new(addr, pwd, method);
                 nsvr.set_mode(global_mode);
 
-                if let Some(p) = config.plugin {
+                if let Some(ref p) = config.plugin {
                     // SIP008 allows "plugin" to be an empty string
                     // Empty string implies "no plugin"
                     if !p.is_empty() {
                         let plugin = PluginConfig {
-                            plugin: p,
-                            plugin_opts: config.plugin_opts,
-                            plugin_args: config.plugin_args.unwrap_or_default(),
+                            plugin: p.clone(),
+                            plugin_opts: config.plugin_opts.clone(),
+                            plugin_args: config.plugin_args.clone().unwrap_or_default(),
                         };
                         nsvr.set_plugin(plugin);
                     }
@@ -1264,7 +1469,12 @@ impl Config {
                             return Err(err);
                         }
                     },
-                    None => nsvr.set_mode(global_mode),
+                    None => {
+                        // Server will derive mode from the global scope
+                        if matches!(config_type, ConfigType::Server | ConfigType::Manager) {
+                            nsvr.set_mode(global_mode);
+                        }
+                    }
                 }
 
                 if let Some(p) = svr.plugin {
@@ -1362,6 +1572,18 @@ impl Config {
                 }
             }
 
+            if let Some(p) = config.plugin {
+                // SIP008 allows "plugin" to be an empty string
+                // Empty string implies "no plugin"
+                if !p.is_empty() {
+                    manager_config.plugin = Some(PluginConfig {
+                        plugin: p,
+                        plugin_opts: config.plugin_opts,
+                        plugin_args: config.plugin_args.unwrap_or_default(),
+                    });
+                }
+            }
+
             nconfig.manager = Some(manager_config);
         }
 
@@ -1407,6 +1629,28 @@ impl Config {
             nconfig.ipv6_first = f;
         }
 
+        // Security
+        if let Some(sec) = config.security {
+            if let Some(replay_attack) = sec.replay_attack {
+                if let Some(policy) = replay_attack.policy {
+                    match policy.parse::<ReplayAttackPolicy>() {
+                        Ok(p) => nconfig.security.replay_attack.policy = p,
+                        Err(..) => {
+                            let err = Error::new(ErrorKind::Invalid, "invalid replay attack policy", None);
+                            return Err(err);
+                        }
+                    }
+                }
+            }
+        }
+
+        if let Some(balancer) = config.balancer {
+            nconfig.balancer = BalancerConfig {
+                max_server_rtt: balancer.max_server_rtt.map(Duration::from_secs),
+                check_interval: balancer.check_interval.map(Duration::from_secs),
+            };
+        }
+
         Ok(nconfig)
     }
 
@@ -1545,11 +1789,19 @@ impl Config {
     }
 
     /// Load Config from a File
-    pub fn load_from_file(filename: &str, config_type: ConfigType) -> Result<Config, Error> {
-        let mut reader = OpenOptions::new().read(true).open(&Path::new(filename))?;
+    pub fn load_from_file<P: AsRef<Path>>(filename: P, config_type: ConfigType) -> Result<Config, Error> {
+        let filename = filename.as_ref();
+
+        let mut reader = OpenOptions::new().read(true).open(filename)?;
         let mut content = String::new();
         reader.read_to_string(&mut content)?;
-        Config::load_from_str(&content[..], config_type)
+
+        let mut config = Config::load_from_str(&content[..], config_type)?;
+
+        // Record the path of the configuration for auto-reloading
+        config.config_path = Some(filename.to_owned());
+
+        Ok(config)
     }
 
     /// Check if there are any plugin are enabled with servers
@@ -1586,6 +1838,21 @@ impl Config {
                 );
                 return Err(err);
             }
+
+            // Balancer related checks
+            if let Some(rtt) = self.balancer.max_server_rtt {
+                if rtt.as_secs() == 0 {
+                    let err = Error::new(ErrorKind::Invalid, "balancer.max_server_rtt must be > 0", None);
+                    return Err(err);
+                }
+            }
+
+            if let Some(intv) = self.balancer.check_interval {
+                if intv.as_secs() == 0 {
+                    let err = Error::new(ErrorKind::Invalid, "balancer.check_interval must be > 0", None);
+                    return Err(err);
+                }
+            }
         }
 
         if self.config_type.is_server() && self.server.is_empty() {
@@ -1659,22 +1926,20 @@ impl fmt::Display for Config {
 
         let mut jconf = SSConfig::default();
 
-        if let Some(ref client) = self.local_addr {
-            jconf.local_address = Some(client.to_string());
-        }
-
         // Locals
         if !self.local.is_empty() {
             if self.local.len() == 1 && self.local[0].is_basic() {
                 let local = &self.local[0];
-                jconf.local_address = Some(match local.addr {
-                    ServerAddr::SocketAddr(ref sa) => sa.ip().to_string(),
-                    ServerAddr::DomainName(ref dm, ..) => dm.to_string(),
-                });
-                jconf.local_port = Some(match local.addr {
-                    ServerAddr::SocketAddr(ref sa) => sa.port(),
-                    ServerAddr::DomainName(.., port) => port,
-                });
+                if let Some(ref a) = local.addr {
+                    jconf.local_address = Some(match a {
+                        ServerAddr::SocketAddr(ref sa) => sa.ip().to_string(),
+                        ServerAddr::DomainName(ref dm, ..) => dm.to_string(),
+                    });
+                    jconf.local_port = Some(match a {
+                        ServerAddr::SocketAddr(ref sa) => sa.port(),
+                        ServerAddr::DomainName(.., port) => *port,
+                    });
+                }
                 if local.protocol != ProtocolType::Socks {
                     jconf.protocol = Some(local.protocol.as_str().to_owned());
                 }
@@ -1682,14 +1947,14 @@ impl fmt::Display for Config {
                 let mut jlocals = Vec::with_capacity(self.local.len());
                 for local in &self.local {
                     let jlocal = SSLocalExtConfig {
-                        local_address: Some(match local.addr {
+                        local_address: local.addr.as_ref().map(|a| match a {
                             ServerAddr::SocketAddr(ref sa) => sa.ip().to_string(),
                             ServerAddr::DomainName(ref dm, ..) => dm.to_string(),
                         }),
-                        local_port: match local.addr {
+                        local_port: local.addr.as_ref().map(|a| match a {
                             ServerAddr::SocketAddr(ref sa) => sa.port(),
-                            ServerAddr::DomainName(.., port) => port,
-                        },
+                            ServerAddr::DomainName(.., port) => *port,
+                        }),
                         disabled: None,
                         local_udp_address: local.udp_addr.as_ref().map(|udp_addr| match udp_addr {
                             ServerAddr::SocketAddr(sa) => sa.ip().to_string(),
@@ -1769,6 +2034,10 @@ impl fmt::Display for Config {
                                 Address::DomainNameAddress(.., port) => Some(*port),
                             },
                         },
+                        #[cfg(feature = "local-tun")]
+                        tun_interface_name: local.tun_interface_name.clone(),
+                        #[cfg(feature = "local-tun")]
+                        tun_interface_address: local.tun_interface_address.as_ref().map(ToString::to_string),
                     };
                     jlocals.push(jlocal);
                 }
@@ -1870,6 +2139,24 @@ impl fmt::Display for Config {
             if jconf.mode.is_none() {
                 jconf.mode = Some(m.mode.to_string());
             }
+
+            if jconf.method.is_none() {
+                if let Some(ref m) = m.method {
+                    jconf.method = Some(m.to_string());
+                }
+            }
+
+            if jconf.plugin.is_none() {
+                if let Some(ref p) = m.plugin {
+                    jconf.plugin = Some(p.plugin.clone());
+                    if let Some(ref o) = p.plugin_opts {
+                        jconf.plugin_opts = Some(o.clone());
+                    }
+                    if !p.plugin_args.is_empty() {
+                        jconf.plugin_args = Some(p.plugin_args.clone());
+                    }
+                }
+            }
         }
 
         if self.no_delay {
@@ -1909,6 +2196,23 @@ impl fmt::Display for Config {
             jconf.ipv6_first = Some(self.ipv6_first);
         }
 
+        // Security
+        if self.security.replay_attack.policy != ReplayAttackPolicy::default() {
+            jconf.security = Some(SSSecurityConfig {
+                replay_attack: Some(SSSecurityReplayAttackConfig {
+                    policy: Some(self.security.replay_attack.policy.to_string()),
+                }),
+            });
+        }
+
+        // Balancer
+        if self.balancer.max_server_rtt.is_some() || self.balancer.check_interval.is_some() {
+            jconf.balancer = Some(SSBalancerConfig {
+                max_server_rtt: self.balancer.max_server_rtt.as_ref().map(Duration::as_secs),
+                check_interval: self.balancer.check_interval.as_ref().map(Duration::as_secs),
+            });
+        }
+
         write!(f, "{}", json5::to_string(&jconf).unwrap())
     }
 }
