@@ -57,9 +57,7 @@ pub async fn run(config: Config) -> io::Result<()> {
         #[cfg(target_os = "android")]
         vpn_protect_path: config.outbound_vpn_protect_path,
 
-        bind_local_addr: config.local_addr,
-
-        #[cfg(any(target_os = "linux", target_os = "android", target_os = "macos", target_os = "ios"))]
+        bind_local_addr: config.outbound_bind_addr,
         bind_interface: config.outbound_bind_interface,
 
         ..Default::default()
@@ -108,6 +106,12 @@ pub async fn run(config: Config) -> io::Result<()> {
             server.set_acl(acl.clone());
         }
 
+        if config.ipv6_first {
+            server.set_ipv6_first(config.ipv6_first);
+        }
+
+        server.set_security_config(&config.security);
+
         servers.push(server);
     }
 
